/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003-2004  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "video.h"
#include "m6809.h"
#include "sam.h"
#include "vdg.h"
#include "pia.h"
#include "logging.h"
#include "types.h"

#include "vdg_bitmaps.c"

static void render_sg4(void);
/* TODO: static void render_sg6(void); */
static void render_cg1(void);
static void render_rg1(void);
static void render_cg2(void);
static void render_rg6(void);
void (*vdg_render_scanline)(void) = render_sg4;

static uint_fast8_t subline;

static Pixel *pixel;
static Pixel darkgreen, black;
static Pixel bg_colour;
static Pixel fg_colour;
static Pixel vdg_colour[8];
static Pixel *cg_colours;
static Pixel border_colour;

void vdg_init(void) {
	vdg_colour[0] = MAPCOLOUR(0x00, 0xff, 0x00);
	vdg_colour[1] = MAPCOLOUR(0xff, 0xff, 0x00);
	vdg_colour[2] = MAPCOLOUR(0x00, 0x00, 0xff);
	vdg_colour[3] = MAPCOLOUR(0xff, 0x00, 0x00);
	vdg_colour[4] = MAPCOLOUR(0xff, 0xe0, 0xe0);
	vdg_colour[5] = MAPCOLOUR(0x00, 0xff, 0xff);
	vdg_colour[6] = MAPCOLOUR(0xff, 0x00, 0xff);
	vdg_colour[7] = MAPCOLOUR(0xff, 0xa5, 0x00);
	black = MAPCOLOUR(0x00, 0x00, 0x00);
	darkgreen = MAPCOLOUR(0x00, 0x20, 0x00);
}

void vdg_reset(void) {
	vdg_set_mode();
	pixel = VIDEO_TOPLEFT + VIDEO_VIEWPORT_YOFFSET;
	subline = 0;
}

void vdg_vsync(void) {
	VIDEO_UPDATE;
	sam_vdg_fsync();
	pixel = VIDEO_TOPLEFT + VIDEO_VIEWPORT_YOFFSET;
	subline = 0;
}

/* Renders a line of alphanumeric/semigraphics 4 (mode is selected by data
 * line, so need to be handled together) */
static void render_sg4(void) {
	uint8_t *vram_ptr;
	uint_fast8_t i, j, octet, tmp;
	for (i = 32; i; i--) {
		*pixel = border_colour;
		*(pixel + 288 * XSTEP) = border_colour;
		pixel += XSTEP;
	}
	for (i = 0; i < 2; i++) {
		vram_ptr = (uint8_t *)sam_vram_ptr(sam_vdg_address);
		for (j = 16; j; j--) {
			octet = *(vram_ptr++);
			if (octet & 0x80) {
				tmp = vdg_colour[(octet & 0x70)>>4];
				if (subline < 6) {
					*pixel = *(pixel+1*XSTEP) = *(pixel+2*XSTEP) = *(pixel+3*XSTEP) = (octet & 0x08) ? tmp : black;
					*(pixel+4*XSTEP) = *(pixel+5*XSTEP) = *(pixel+6*XSTEP) = *(pixel+7*XSTEP) = (octet & 0x04) ? tmp : black;
				} else {
					*pixel = *(pixel+1*XSTEP) = *(pixel+2*XSTEP) = *(pixel+3*XSTEP) = (octet & 0x02) ? tmp : black;
					*(pixel+4*XSTEP) = *(pixel+5*XSTEP) = *(pixel+6*XSTEP) = *(pixel+7*XSTEP) = (octet & 0x01) ? tmp : black;
				}
				pixel += 8*XSTEP;
			} else {
				tmp = vdg_alpha[(octet&0x3f)*12 + subline];
				if (octet & 0x40)
					tmp = ~tmp;
				*pixel = (tmp & 0x80) ? fg_colour : bg_colour;
				*(pixel+1*XSTEP) = (tmp & 0x40) ? fg_colour : bg_colour;
				*(pixel+2*XSTEP) = (tmp & 0x20) ? fg_colour : bg_colour;
				*(pixel+3*XSTEP) = (tmp & 0x10) ? fg_colour : bg_colour;
				*(pixel+4*XSTEP) = (tmp & 0x08) ? fg_colour : bg_colour;
				*(pixel+5*XSTEP) = (tmp & 0x04) ? fg_colour : bg_colour;
				*(pixel+6*XSTEP) = (tmp & 0x02) ? fg_colour : bg_colour;
				*(pixel+7*XSTEP) = (tmp & 0x01) ? fg_colour : bg_colour;
				pixel += 8*XSTEP;
			}
		}
		sam_vdg_xstep(16);
	}
	sam_vdg_hsync(32,10,16);
	pixel += NEXTLINE + (32 * XSTEP);
	subline++;
	if (subline > 11)
		subline = 0;
}

#define RENDER_BYTE_CG1(b,o,n) *(pixel+(o)*XSTEP) = *(pixel+(o+1)*XSTEP) \
		= *(pixel+(o+2)*XSTEP) = *(pixel+(o+3)*XSTEP) \
		= cg_colours[(b & (0xc0<<(n * 8))) >> ((n*8)+6)]; \
	*(pixel+(o+4)*XSTEP) = *(pixel+(o+5)*XSTEP) \
		= *(pixel+(o+6)*XSTEP) = *(pixel+(o+7)*XSTEP) \
		= cg_colours[(b & (0x30<<(n * 8))) >> ((n*8)+4)]; \
	*(pixel+(o+8)*XSTEP) = *(pixel+(o+9)*XSTEP) \
		= *(pixel+(o+10)*XSTEP) = *(pixel+(o+11)*XSTEP) \
		= cg_colours[(b & (0x0c<<(n * 8))) >> ((n*8)+2)]; \
	*(pixel+(o+12)*XSTEP) = *(pixel+(o+13)*XSTEP) \
		= *(pixel+(o+14)*XSTEP) = *(pixel+(o+15)*XSTEP) \
		= cg_colours[(b & (0x03<<(n * 8))) >> ((n*8))]

/* Render a 16-byte colour graphics line (CG1) */
static void render_cg1(void) {
	uint32_t *vram_ptr;
	uint_fast8_t i;
	uint_fast32_t octet;
	for (i = 32; i; i--) {
		*pixel = border_colour;
		*(pixel + 288 * XSTEP) = border_colour;
		pixel += XSTEP;
	}
	vram_ptr = (uint32_t *)sam_vram_ptr(sam_vdg_address);
	for (i = 4; i; i--) {
		octet = *(vram_ptr++);
#ifdef WRONG_ENDIAN
		RENDER_BYTE_CG1(octet,0,0);
		RENDER_BYTE_CG1(octet,16,1);
		RENDER_BYTE_CG1(octet,32,2);
		RENDER_BYTE_CG1(octet,48,3);
#else
		RENDER_BYTE_CG1(octet,0,3);
		RENDER_BYTE_CG1(octet,16,2);
		RENDER_BYTE_CG1(octet,32,1);
		RENDER_BYTE_CG1(octet,48,0);
#endif
		pixel += 64*XSTEP;
	}
	sam_vdg_xstep(16);
	sam_vdg_hsync(16,6,0);
	pixel += NEXTLINE + (32 * XSTEP);
	subline++;
	if (subline > 11)
		subline = 0;
}

#define RENDER_BYTE_RG1(b,o,n) *(pixel+(o)*XSTEP) = *(pixel+(o+1)*XSTEP) = (b & (0x80<<(n*8))) ? fg_colour : bg_colour; \
	*(pixel+(o+2)*XSTEP) = *(pixel+(o+3)*XSTEP) = (b & (0x40<<(n*8))) ? fg_colour : bg_colour; \
	*(pixel+(o+4)*XSTEP) = *(pixel+(o+5)*XSTEP) = (b & (0x20<<(n*8))) ? fg_colour : bg_colour; \
	*(pixel+(o+6)*XSTEP) = *(pixel+(o+7)*XSTEP) = (b & (0x10<<(n*8))) ? fg_colour : bg_colour; \
	*(pixel+(o+8)*XSTEP) = *(pixel+(o+9)*XSTEP) = (b & (0x08<<(n*8))) ? fg_colour : bg_colour; \
	*(pixel+(o+10)*XSTEP) = *(pixel+(o+11)*XSTEP) = (b & (0x04<<(n*8))) ? fg_colour : bg_colour; \
	*(pixel+(o+12)*XSTEP) = *(pixel+(o+13)*XSTEP) = (b & (0x02<<(n*8))) ? fg_colour : bg_colour; \
	*(pixel+(o+14)*XSTEP) = *(pixel+(o+15)*XSTEP) = (b & (0x01<<(n*8))) ? fg_colour : bg_colour

/* Render a 16-byte resolution graphics line (RG1,RG2,RG3) */
static void render_rg1(void) {
	uint32_t *vram_ptr;
	uint_fast8_t i;
	uint_fast32_t octet;
	for (i = 32; i; i--) {
		*pixel = border_colour;
		*(pixel + 288 * XSTEP) = border_colour;
		pixel += XSTEP;
	}
	vram_ptr = (uint32_t *)sam_vram_ptr(sam_vdg_address);
	for (i = 4; i; i--) {
		octet = *(vram_ptr++);
#ifdef WRONG_ENDIAN
			RENDER_BYTE_RG1(octet,0,0);
			RENDER_BYTE_RG1(octet,16,1);
			RENDER_BYTE_RG1(octet,32,2);
			RENDER_BYTE_RG1(octet,48,3);
#else
			RENDER_BYTE_RG1(octet,0,3);
			RENDER_BYTE_RG1(octet,16,2);
			RENDER_BYTE_RG1(octet,32,1);
			RENDER_BYTE_RG1(octet,48,0);
#endif
		pixel += 64*XSTEP;
	}
	sam_vdg_xstep(16);
	sam_vdg_hsync(16,6,0);
	pixel += NEXTLINE + (32 * XSTEP);
	subline++;
	if (subline > 11)
		subline = 0;
}

#define RENDER_BYTE_CG2(b,o,n) *(pixel+(o)*XSTEP) = *(pixel+(o+1)*XSTEP) = cg_colours[(b & (0xc0<<(n * 8))) >> ((n*8)+6)]; \
	*(pixel+(o+2)*XSTEP) = *(pixel+(o+3)*XSTEP) = cg_colours[(b & (0x30<<(n * 8))) >> ((n*8)+4)]; \
	*(pixel+(o+4)*XSTEP) = *(pixel+(o+5)*XSTEP) = cg_colours[(b & (0x0c<<(n * 8))) >> ((n*8)+2)]; \
	*(pixel+(o+6)*XSTEP) = *(pixel+(o+7)*XSTEP) = cg_colours[(b & (0x03<<(n * 8))) >> ((n*8))]

/* Render a 32-byte colour graphics line (CG2,CG3,CG6) */
static void render_cg2(void) {
	uint32_t *vram_ptr;
	uint_fast8_t i, j;
	uint_fast32_t octet;
	for (i = 32; i; i--) {
		*pixel = border_colour;
		*(pixel + 288 * XSTEP) = border_colour;
		pixel += XSTEP;
	}
	for (i = 0; i < 2; i++) {
		vram_ptr = (uint32_t *)sam_vram_ptr(sam_vdg_address);
		for (j = 4; j; j--) {
			octet = *(vram_ptr++);
#ifdef WRONG_ENDIAN
			RENDER_BYTE_CG2(octet,0,0);
			RENDER_BYTE_CG2(octet,8,1);
			RENDER_BYTE_CG2(octet,16,2);
			RENDER_BYTE_CG2(octet,24,3);
#else
			RENDER_BYTE_CG2(octet,0,3);
			RENDER_BYTE_CG2(octet,8,2);
			RENDER_BYTE_CG2(octet,16,1);
			RENDER_BYTE_CG2(octet,24,0);
#endif
			pixel += 32*XSTEP;
		}
		sam_vdg_xstep(16);
	}
	sam_vdg_hsync(32,10,16);
	pixel += NEXTLINE + (32 * XSTEP);
	subline++;
	if (subline > 11)
		subline = 0;
}

#define RENDER_BYTE_RG6(b,o,n) *(pixel+(o)*XSTEP) = (b & (0x80<<(n*8))) ? fg_colour : bg_colour; \
	*(pixel+(o+1)*XSTEP) = (b & (0x40<<(n*8))) ? fg_colour : bg_colour; \
	*(pixel+(o+2)*XSTEP) = (b & (0x20<<(n*8))) ? fg_colour : bg_colour; \
	*(pixel+(o+3)*XSTEP) = (b & (0x10<<(n*8))) ? fg_colour : bg_colour; \
	*(pixel+(o+4)*XSTEP) = (b & (0x08<<(n*8))) ? fg_colour : bg_colour; \
	*(pixel+(o+5)*XSTEP) = (b & (0x04<<(n*8))) ? fg_colour : bg_colour; \
	*(pixel+(o+6)*XSTEP) = (b & (0x02<<(n*8))) ? fg_colour : bg_colour; \
	*(pixel+(o+7)*XSTEP) = (b & (0x01<<(n*8))) ? fg_colour : bg_colour

/* Render a 32-byte resolution graphics line (RG6) */
static void render_rg6(void) {
	uint32_t *vram_ptr;
	uint_fast8_t i, j;
	uint_fast32_t octet;
	for (i = 32; i; i--) {
		*pixel = border_colour;
		*(pixel + 288 * XSTEP) = border_colour;
		pixel += XSTEP;
	}
	for (i = 0; i < 2; i++) {
		vram_ptr = (uint32_t *)sam_vram_ptr(sam_vdg_address);
		for (j = 4; j; j--) {
			octet = *(vram_ptr++);
#ifdef WRONG_ENDIAN
			RENDER_BYTE_RG6(octet,0,0);
			RENDER_BYTE_RG6(octet,8,1);
			RENDER_BYTE_RG6(octet,16,2);
			RENDER_BYTE_RG6(octet,24,3);
#else
			RENDER_BYTE_RG6(octet,0,3);
			RENDER_BYTE_RG6(octet,8,2);
			RENDER_BYTE_RG6(octet,16,1);
			RENDER_BYTE_RG6(octet,24,0);
#endif
			pixel += 32*XSTEP;
		}
		sam_vdg_xstep(16);
	}
	sam_vdg_hsync(32,10,16);
	pixel += NEXTLINE + (32 * XSTEP);
	subline++;
	if (subline > 11)
		subline = 0;
}

void vdg_set_mode(void) {
	uint_fast8_t mode = PIA_1B.port_output;
	if (mode & 0x80) {
		/* Graphics modes */
		cg_colours = &vdg_colour[(mode & 0x08) >> 1];
		bg_colour = black;
		fg_colour = cg_colours[0];
		border_colour = fg_colour;
		switch ((mode & 0x70) >> 4) {
			case 0: vdg_render_scanline = render_cg1; break;
			case 1:
			case 3:
			case 5: vdg_render_scanline = render_rg1; break;
			case 2:
			case 4:
			case 6: vdg_render_scanline = render_cg2; break;
			default:
			case 7: vdg_render_scanline = render_rg6; break;
		}
	} else {
		bg_colour = darkgreen;
		border_colour = black;
		if (mode & 0x08)
			fg_colour = vdg_colour[7];
		else
			fg_colour = vdg_colour[0];
		vdg_render_scanline = render_sg4;
	}
}
