/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003-2004  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>

#include "video.h"
#include "types.h"

static Pixel backup[320*240];

GPDRAWSURFACE screen;

void video_init(void) {
	GpGraphicModeSet(VIDEO_BPP, NULL);
	GpLcdSurfaceGet(&screen, 0);
	GpSurfaceSet(&screen);
	GpLcdEnable();
	swi_mmu_change((unsigned char *)((uint32_t)screen.ptbuffer & ~4095), (unsigned char *)((uint32_t)(screen.ptbuffer + (320 * 240 * sizeof(Pixel)) + 4095) & ~4095), MMU_NCNB);
	video_fillrect(0,0,320,196,MAPCOLOUR(0,0,0));
	video_fillrect(0,196,320,44,MAPCOLOUR(255,255,255));
}

void video_shutdown(void) {
}

void video_fillrect(uint_fast16_t x, uint_fast16_t y, uint_fast16_t w, uint_fast16_t h, Pixel colour) {
	Pixel *p = (Pixel *)screen.ptbuffer + x * 240 + (239-y);
	uint_fast16_t skip = 240 + h;
	uint_fast16_t j;
	for (; w; w--) {
		for (j = h; j; j--) {
			*(p--) = colour;
		}
		p += skip;
	}
}

void video_blit(uint_fast16_t x, uint_fast16_t y, Sprite *src) {
	Pixel *s = src->data;
	Pixel *d = VIDEO_SCREENBASE + (x * 240) + (240-y) - src->h;
	uint_fast16_t j, w = src->w;
	for (; w; w--) {
		for (j = src->h; j; j--) {
			*(d++) = *(s++);
		}
		d += (240 - src->h);
	}
}

void video_backup(void) {
	memcpy(backup, VIDEO_SCREENBASE, 320*240*sizeof(Pixel));
}

void video_restore(void) {
	memcpy(VIDEO_SCREENBASE, backup, 320*240*sizeof(Pixel));
}
