/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003-2004  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>

#include "video.h"
#include "types.h"
#include "logging.h"

static Pixel backup[320*240];

SDL_Surface *screen;

void video_init(void) {
	LOG_DEBUG(2,"Initialising SDL video driver\n");
	if (SDL_InitSubSystem(SDL_INIT_VIDEO) < 0) {
		LOG_ERROR("Failed to initialiase SDL video driver\n");
		exit(1);
	}
	screen = SDL_SetVideoMode(320, 240, 8, SDL_SWSURFACE);
	if (screen == NULL) {
		LOG_ERROR("Failed to allocate SDL surface for display\n");
		exit(1);
	}
	video_fillrect(0,0,320,240,MAPCOLOUR(0,0,0));
}

void video_shutdown(void) {
	LOG_DEBUG(2,"Shutting down SDL video driver\n");
	SDL_QuitSubSystem(SDL_INIT_VIDEO);
}

void video_fillrect(uint_fast16_t x, uint_fast16_t y, uint_fast16_t w, uint_fast16_t h, Pixel colour) {
	Pixel *p1 = VIDEO_SCREENBASE + (239 - y) * 320 + x;
	int skip = -320 - w;
	uint_fast16_t i;
	for (; h; h--) {
		for (i = w; i; i--) {
			*(p1++) = colour;
		}
		p1 += skip;
	}
}

void video_blit(uint_fast16_t x, uint_fast16_t y, Sprite *src);

void video_backup(void) {
	memcpy(backup, VIDEO_SCREENBASE, 320*240*sizeof(Pixel));
}

void video_restore(void) {
	memcpy(VIDEO_SCREENBASE, backup, 320*240*sizeof(Pixel));
}
