/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003-2004  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "xroar.h"
#include "video.h"
#include "sound.h"
#include "keyboard.h"
#include "joystick.h"
#include "ui.h"
#include "fs.h"
#include "machine.h"
#include "m6809.h"
#include "pia.h"
#include "wd2797.h"
#include "snapshot.h"
#include "logging.h"

Cycle current_cycle;
static Cycle next_hsync, next_keyboard_poll;
Cycle next_sound_update, next_disk_interrupt;
int_fast8_t enable_disk_interrupt;
static int_fast16_t vdg_call;

#ifdef TRACE
int trace = 0;
#endif

void xroar_init(void) {
	/* Initialise everything */
	video_init();
	sound_init();
	keyboard_init();
	joystick_init();
	ui_init();
	fs_init();
	current_cycle = 0;
	machine_init();
	xroar_reset(RESET_HARD);
}

void xroar_shutdown(void) {
	joystick_shutdown();
	video_shutdown();
	sound_shutdown();
}

void xroar_reset(int hard) {
	vdg_call = 0;
	enable_disk_interrupt = 0;
	next_hsync = current_cycle + CYCLES_PER_SCANLINE;
	next_keyboard_poll = current_cycle + 141050;
	/* Reset everything */
	sound_reset();
	joystick_reset();
	keyboard_reset();
	machine_reset(hard);
}

void xroar_mainloop(void) {
#ifdef AUTOSNAP_HACK
	snapshot_state = 2;  /* XXX: testing */
#endif
	while (1) {
		Cycle until;
		if (snapshot_state) {
			switch(snapshot_state) {
				case 1: write_snapshot(); break;
				case 2: read_snapshot(); break;
				case 3: xroar_reset(RESET_SOFT); break;
				case 4: xroar_reset(RESET_HARD); break;
			}
			snapshot_state = 0;
		}
		if ((int)(current_cycle - next_hsync) >= 0) {
			next_hsync += CYCLES_PER_SCANLINE;
			if ((vdg_call >= TOP_BORDER_OFFSET) && (vdg_call < (TOP_BORDER_OFFSET+192)))
#ifdef GP32
				if ((vdg_call & 3) == 0)
#endif
					vdg_render_scanline();
			PIA_SET_P0CA1;
			vdg_call++;
			if (vdg_call == ACTIVE_SCANLINES_PER_FRAME) {
				vdg_vsync();
				PIA_SET_P0CB1;
			}
			if (vdg_call >= TOTAL_SCANLINES_PER_FRAME) {
				vdg_call = 0;
			}
		}
		if ((int)(current_cycle - next_keyboard_poll) >= 0) {
			next_keyboard_poll += 141050;
			keyboard_poll();
			joystick_poll();
		}
		if ((int)(current_cycle - next_sound_update) >= 0) {
			sound_update();
		}
		if (enable_disk_interrupt && (int)(current_cycle - next_disk_interrupt) >= 0) {
			wd2797_generate_interrupt();
		}
		if (next_hsync < next_sound_update)
			until = next_hsync;
		else
			until = next_sound_update;
		if (next_keyboard_poll < until) until = next_keyboard_poll;
		if (enable_disk_interrupt && next_disk_interrupt < until)
			until = next_disk_interrupt;
		m6809_cycle(until);
	}
}
